function denoised = NLMSBW(image, params, stdNoise, cleanImage)
    % Main function of the NLM-SBW denoising algorithm.
    % INPUT ARGUMENTS:
    % image
    %   A corrupted grayscale image.
    % SNR
    %   The signal-to-noise ratio of the image.
    % OUTPUT ARGUMENTS:
    % denoised
    %   The denoised image.    

    if ~isstruct(params)        
        params = DefaultParams(params);            
    end
    
    %image = ClampImage(image);
    features = GetImagePatches(image, params.alg.patchSize);
    
    mask = ones(size(image));    
    
    numPotentialCentroids = numel(mask);    
    
    clusterPix = {};
    clusterInd = {};
    clusterDist = {};
    
    iterationCounter = 1;
    while numPotentialCentroids > 0
        [ curClusterInd, curClusterDist ] = RandomCluster(image, mask, features, params.alg.sigma);        
        
        denoisedCluster = DenoiseCluster(image(curClusterInd), features(:, curClusterInd), params.alg);                        
        
%         if numel(curClusterInd) > 1000
%             [ denoisedCluster ]  = DenoiseClusterSure(image, curClusterInd, params.alg, stdNoise, features(:, curClusterInd), cleanImage);        
%         else
%             denoisedCluster = DenoiseCluster(image(curClusterInd), features(:, curClusterInd), params.alg);                
%         end
        
        
        clusterPix{iterationCounter} = denoisedCluster;
        clusterInd{iterationCounter} = curClusterInd;
        clusterDist{iterationCounter} = curClusterDist;
        
        mask(curClusterInd) = 0;
        prevNumPotentialCentroids = numPotentialCentroids;
        numPotentialCentroids = numel(find(mask));                
        iterationCounter = iterationCounter + 1;
        
        DEBUG = 1;
        if DEBUG
            fprintf(1, 'iteration=%d | numPotentialCentroids=%d | Finished=%d | Cluster size=%d\n', ...
            iterationCounter, numPotentialCentroids, prevNumPotentialCentroids - numPotentialCentroids, numel(curClusterInd)); 
        
            SHOW_BLOCKS = 0;
            if SHOW_BLOCKS
                curDenoisedBlock = zeros(size(image));
                curDenoisedBlock(curClusterInd) = denoisedCluster;
                imshow(curDenoisedBlock);
                pause;
            end
        end
    end        
    
    blended = BlendClusters(size(image), clusterPix, clusterInd, clusterDist);
    denoised = ClampImage(blended);
end

function clamped = ClampImage(image)
    clamped = max(0, min(image, 1));
end

function blended = BlendClusters(imageSize, clusterPix, clusterInd, clusterDist)
    blended = zeros(imageSize);
    clusterWeights = zeros(imageSize);
    for i = 1 : numel(clusterPix)
        %curWeights = 1 ./ (1 + clusterDist{i});
        curWeights = exp(-clusterDist{i} / 10);
        blended(clusterInd{i}) = blended(clusterInd{i}) + curWeights .* clusterPix{i};
        clusterWeights(clusterInd{i}) = clusterWeights(clusterInd{i}) + curWeights;
    end
    blended = blended ./ clusterWeights;
end

function [ subclusterInd, clusterDist ] = RandomCluster(image, mask, features, sigma)    
    minAffThresh = 0.001;
    maxClusterSize = 4000;
    sigma =  sigma * 1.25;
    patchSim = @(x,y)(exp(-sum((x - y).^2) ./ sigma^2));
    
    
    maskInd = find(mask);
    clusterCentroidInd = maskInd(randi(numel(maskInd)));
    [centroidRow, centroidCol] = ind2sub(size(image), clusterCentroidInd);   
    
    [ X, Y ] = meshgrid(1 : size(image, 2), 1 : size(image, 1));
    %distances = max(abs(X - centroidCol), abs(Y-centroidRow));
    distances = sqrt((X - centroidCol).^2 + (Y-centroidRow).^2);    
    
    MAX_RAD = 30;
    indToCheck = find(distances < MAX_RAD)';
    
    sim = zeros(numel(image), 1);
    
    featuresSqrDist = sqdistance(features(:, indToCheck), features(:, clusterCentroidInd));
    featuresExp = exp(-featuresSqrDist ./ sigma^2);
    sim(indToCheck) = featuresExp;    
    
    clusterInd = find(sim > minAffThresh);  
    
    %clusterInd = intersect(clusterInd, maskInd);
    %numCoveredCentroids = numel(intersect(clusterInd, maskInd));    
    %if numCoveredCentroids < 5
    %    maxClusterSize = min(maxClusterSize, max(numCoveredCentroids, 200));
    %end
    
    subclusterInd = SubCluster(size(mask), clusterInd, distances, maxClusterSize);
    
    %[ simSorted sortInd ] = sort(sim(clusterInd), 'descend');
    %numPatches = min(numel(simSorted), maxClusterSize);
    %subclusterInd = clusterInd(sortInd(1 : numPatches));
    
    clusterDist = distances(subclusterInd);
    
     if 0
        result = zeros(size(mask));
        result(clusterInd) = 1;        
        imshow([image, result .* image, result]);
        pause;
    end
end

function subclusterInd = SubCluster(imageSize, clusterInd, distances, maxClusterSize)
    clusterMask = zeros(imageSize);
    clusterMask(clusterInd) = 1;    
    
    distMask = (2*clusterMask-1) .* distances;
    
    curSubClusterSize = 0;
    curDist = 1;    
    %numIter = 1;
    %maxNumIter = 100;
    while curSubClusterSize < maxClusterSize
        curDist = ceil(curDist * 1.2);
        subclusterInd = find(distMask <= curDist & distMask >= 0);
        preSubClusterSize  = curSubClusterSize;
        curSubClusterSize = numel(subclusterInd);        
        %numIter = numIter + 1;
        %if numIter >= maxNumIter
        %    break
        %end
        if preSubClusterSize == curSubClusterSize
            break
        end
    end
    
    %subclusterInd =  [ subclusterInd; centroidInd ];
    
    if 0
        imshow(clusterMask)
        figure
        result = zeros(size(clusterMask));
        result(subclusterInd) = 1;
        imshow(result);
    end
end

function denoised = DenoiseCluster(cluster, features, params)
    % Denoises a single image block.
    % INPUT ARGUMENTS:
    % imageBlock
    %   A corrupted grayscale image.
    % params
    %   A structure of NLM-SBW parameters.
    % OUTPUT ARGUMENTS:
    % denoised
    %   The denoised image block.
        
    W = BuildWeightsMatrix(features, params.sigma);
    A = RowNormalize(W);  
    clear W;
    
    coef = NRChebychev(params.bwCutoff, params.bwOrder, params.numChebyCoef);
    denoised = ChebychevClenshawMatVec(A, coef, 0, 1, cluster(:));
    %denoised = A * cluster(:);
end

function [ denoisedSure ] = DenoiseClusterSure(image, clusterInd, params, stdNoise, featuresOrig, cleanImage)    
    
    cutoffs = [  0.6 0.7 0.8  ];
    %cutoffs = [ 5 10 20 30 50 ];
    numIter = 3;
    
    sureEps = 10e-14;
    
    mseEstimate = zeros(numel(cutoffs), 1);    
    
    denoisedVariants = zeros(numel(cutoffs), numel(clusterInd));
    
    for i = 1 : numel(cutoffs)
        curParams = params;
        %curParams.bwCutoff = cutoffs(i);
        curParams.bwOrder = cutoffs(i);
        
        coef = NRChebychev(curParams.bwCutoff, curParams.bwOrder, curParams.numChebyCoef);
        
        denoisedBase = DenoiseCluster(image(clusterInd), featuresOrig, curParams);                
        denoisedVariants(i, :, :) = denoisedBase;
        
        divInst = zeros(numIter, 1);        
        for j = 1 : numIter            
            b = randn(size(image));
            perturbed = image + sureEps * b;            
            
            features = GetImagePatchesByInd(perturbed, clusterInd, curParams.patchSize);
            
            W = BuildWeightsMatrix(features, curParams.sigma);
            A = RowNormalize(W);  
            clear W;
            
            cluster = perturbed(clusterInd);
            denoisedPerturbed = ChebychevClenshawMatVec(A, coef, 0, 1, cluster(:));                        
            
            divInst(j) = (1 / sureEps) * sum(sum(b(clusterInd).*(denoisedPerturbed - denoisedBase)));
        end                
        divergence = mean(divInst(divInst ~= 0));
        N = numel(clusterInd);    
        mseEstimate(i) = (-N*stdNoise^2 + sum(sum((denoisedBase - image(clusterInd)).^2)) + 2 * stdNoise^2 * divergence) / N;
        mseTrue(i) = mean((denoisedBase - cleanImage(clusterInd)).^2);
    end  
    
    DEBUG = 0;
    if DEBUG 
        plot(cutoffs, mseTrue)
        hold on; plot(cutoffs, mseEstimate, 'r')
        pause;
        close all;
    end
    
    [ ~, minTrueInd ] = min(mseTrue);
    [ ~, minInd ] = min(mseEstimate);
    
    if 1%minTrueInd ~= minInd
        fprintf('true %d estimate %d\n', minTrueInd, minInd);
    end
    
    denoisedSure = denoisedVariants(minInd, :, :)';    
end

function features = GetImagePatchesByInd(image, pixInd, patchDiameter)       
    features = zeros(patchDiameter ^ 2, numel(pixInd));
    
    padSize = (patchDiameter - 1)  / 2;
    imPadded = padarray(image, [ padSize, padSize], 'symmetric');
    
    [ row col ] = ind2sub(size(image), pixInd);
    for i = 1 : numel(pixInd)    
        rowInd = row(i) : (row(i) + patchDiameter - 1);
        colInd = col(i) : (col(i) + patchDiameter - 1);
        curPatch = imPadded(rowInd, colInd);
        features(:, i) = curPatch(:);
    end
end

function features = GetImagePatches(im, patchDiameter)
    % Returns a feature matrix consisting of square patches
    % extracted from the image.  
    %
    % INPUT ARGUMENTS:
    %   im            
    %       MxN image.
    %   patchdiameter 
    %       length of square side.
    %
    % OUTPUT ARGUMENTS:
    %   features    
    %       Matrix of dimensions(patchdiameter^2) x (M*N).

    [M N] = size(im);
    K = patchDiameter^2;

    % construct convolution matrices
    features = zeros(M, N, K);
    index = 1;

    for j=1:patchDiameter
        for i=1:patchDiameter    

            C = zeros(patchDiameter);
            C(i,j) = 1.0;

            features(:,:, index) = imfilter(im, C, 'symmetric', 'same', 'conv');
            index = index+1;
        end
    end

    features = reshape(features, [M*N, K]);
    features = features';
    features = flipud(features);
end

function weights = BuildWeightsMatrix(features, sigma)
    % Builds the NLM weights matrix.
    % INPUT ARGUMENTS:
    % features
    %   The feature matrix.
    % sigma
    %   The NLM kernel width.
    % OUTPUT ARGUMENTS:
    % params
    %   See below for parameter descriptions.
    weights = sqdistance(features);    
    weights = (exp(-(weights)./(sigma.^2)));
    
    %nondiagratio = sum(weights - diag(diag(weights)), 2) ./ sum(weights, 2);

    %disp(min(nondiagratio));
end

function normalized = RowNormalize(mat)
    % Performs row-normalization on a matrix.
    % INPUT ARGUMENTS:
    % mat
    %   The matrix to normalized.
    % OUTPUT ARGUMENTS:
    % mat
    %   The input matrix, transformed such that each row will sum to unity.
    D = sum(mat,2);
    Db = D(:, ones(1, size(mat, 2)));
    normalized = mat ./ Db;
end

function [ y ] = ChebychevClenshawMatVec(x, coef, a, b, v)
    % Computes a the product of a Chebychev polynomial of a matrix and a column vector.
    % INPUT ARGUMENTS:
    % x
    %   The matrix.
    % coef
    %   The Chebychev coefficients.
    % a, b
    %   The range on which the polynomial is defined.
    % v
    %   The column vector by which the matrix polynomial is to be multiplied.
    % OUTPUT ARGUMENTS:
    % y
    %   The column vector resulting from multiplying the polynomial of the matrix x by the
    %   input vector v.

%     n = numel(coef);
%     dim = size(x, 1);    
%     t = 2 * (x - a) ./ (b - a);
%     t(1:dim+1:dim*dim) = t(1:dim+1:dim*dim) - 1;
%     d = zeros(dim, 1);
%     dd = zeros(dim, 1);
%     for i = n : - 1 : 2
%         sv = d;
%         d = 2 * (t * d) - dd + coef(i) * v;
%         dd = sv;
%     end
%     y = t * d - dd + 0.5 * coef(1) * v;

    n = numel(coef);
    dim = size(x, 1);
    I = eye(dim);
    t = 2 * (x - a) ./ (b - a) - I;
    d = zeros(dim, 1);
    dd = zeros(dim, 1);
    for i = n : - 1 : 2
        sv = d;
        %d = 2 * t * d - dd + coef(i) * I * v;
        d = 2 * (t * d) - dd + coef(i) * v;
        dd = sv;
    end
    y = t * d - dd + 0.5 * coef(1) * I * v;
end